"""Error types that methods in the FaunaDB client throw."""

# pylint: disable=redefined-builtin

from builtins import object

from typing import Union

from pydantic import ValidationError

from aiohttp.web_exceptions import HTTPException


class codes:

    """HTTP status codes."""

    bad_request = 400

    unauthorized = 401

    forbidden = 403

    not_found = 404

    internal_server_error = 500

    unavailable = 503


def _get_or_raise(request_result, dct, key):

    if isinstance(dct, dict) and key in dct:

        return dct[key]

    else:

        raise UnexpectedError(
            "Response JSON does not contain expected key %s" % key, request_result
        )


# region FaunaError


class FaunaError(Exception):
    """

    Error returned by the FaunaDB server.

    For documentation of error types, see the `docs <https://fauna.com/documentation#errors>`__.
    """

    @staticmethod
    def raise_for_status_code(request_result):

        code = request_result.status_code

        # pylint: disable=no-member, too-many-return-statements

        if 200 <= code <= 299:
            pass

        elif code == codes.bad_request:

            raise BadRequest(request_result)

        elif code == codes.unauthorized:

            raise Unauthorized(request_result)

        elif code == codes.forbidden:

            raise PermissionDenied(request_result)

        elif code == codes.not_found:

            raise NotFound(request_result)

        elif code == codes.internal_server_error:

            raise InternalError(request_result)

        elif code == codes.unavailable:

            raise UnavailableError(request_result)

        else:

            raise UnexpectedError("Unexpected status code.", request_result)

    def __init__(self, description, request_result):

        super(FaunaError, self).__init__(description)

        self.request_result = request_result

        """:any:`RequestResult` for the request that caused this error."""


class UnexpectedError(FaunaError):

    """Error for when the server returns an unexpected kind of response."""

    pass


class HttpError(FaunaError):
    def __init__(self, request_result):

        self.errors = HttpError._get_errors(request_result)

        """List of all :py:class:`ErrorData` objects sent by the server."""

        super(HttpError, self).__init__(self._get_description(), request_result)

    @staticmethod
    def _get_errors(request_result):

        response = request_result.response_content

        errors = _get_or_raise(request_result, response, "errors")

        return [ErrorData.from_dict(error, request_result) for error in errors]

    def __str__(self):

        return repr(self.errors[0])

    def _get_description(self):

        return self.errors[0].description if self.errors else "(empty `errors`)"


class BadRequest(HttpError):

    """HTTP 400 error."""

    pass


class Unauthorized(HttpError):
    def __init__(self, request_result):

        super(Unauthorized, self).__init__(request_result)

        self.errors[
            0
        ].description = "Unauthorized. Check that endpoint, schema, port and secret are correct during client’s instantiation"


class PermissionDenied(HttpError):

    """HTTP 403 error."""

    pass


class NotFound(HttpError):

    """HTTP 404 error."""

    pass


class InternalError(HttpError):

    """HTTP 500 error."""

    pass


class UnavailableError(HttpError):

    """HTTP 503 error."""

    pass


# endregion


class ErrorData(object):
    """

    Data for one error returned by the server.
    """

    @staticmethod
    def from_dict(dct, request_result):

        return ErrorData(
            _get_or_raise(request_result, dct, "code"),
            _get_or_raise(request_result, dct, "description"),
            dct.get("position"),
            ErrorData.get_failures(dct, request_result),
            ErrorData.get_cause(dct, request_result),
        )

    @staticmethod
    def get_failures(dct, request_result):

        if "failures" in dct:

            return [
                Failure.from_dict(failure, request_result)
                for failure in dct["failures"]
            ]

        return None

    @staticmethod
    def get_cause(dct, request_result):

        if "cause" in dct:

            return [
                ErrorData.from_dict(cause, request_result) for cause in dct["cause"]
            ]

        return None

    def __init__(self, code, description, position, failures, cause=None):

        self.code = code

        """Error code. See all error codes `here <https://fauna.com/documentation#errors>`__."""

        self.description = description

        """Error description."""

        self.position = position

        """Position of the error in a query. May be None."""

        self.failures = failures

        """Cause of the error. May be None."""

        self.cause = cause
        """

    List of all :py:class:`Failure` objects returned by the server.

    None unless code == "validation failed".
    """

    def __repr__(self):

        return (
            "ErrorData(code=%s, description=%s, position=%s, failures=%s, cause=%s)"
            % (
                repr(self.code),
                repr(self.description),
                repr(self.position),
                repr(self.failures),
                repr(self.cause),
            )
        )

    def __eq__(self, other):

        return (
            self.__class__ == other.__class__
            and self.description == other.description
            and self.position == other.position
            and self.failures == other.failures
            and self.cause == other.cause
        )

    def __ne__(self, other):

        # pylint: disable=unneeded-not

        return not self == other


class Failure(object):
    """

    Part of the ``failures`` of an :py:class:`ErrorData`.

    See the ``Invalid Data`` section of the `docs <https://fauna.com/documentation#errors>`__.
    """

    @staticmethod
    def from_dict(dct, request_result):

        return Failure(
            _get_or_raise(request_result, dct, "code"),
            _get_or_raise(request_result, dct, "description"),
            _get_or_raise(request_result, dct, "field"),
        )

    def __init__(self, code, description, field):

        self.code = code

        """Failure code."""

        self.description = description

        """Failure description."""

        self.field = field

        """Field of the failure in the instance."""

    def __repr__(self):

        return "Failure(code=%s, description=%s, field=%s)" % (
            repr(self.code),
            repr(self.description),
            repr(self.field),
        )

    def __eq__(self, other):

        return (
            self.code == other.code
            and self.description == other.description
            and self.field == other.field
        )

    def __ne__(self, other):

        # pylint: disable=unneeded-not

        return not self == other


AioFaunaException = (
    ValidationError,
    FaunaError,
    HTTPException,
    BaseException,
    Exception,
    TypeError,
    ValueError,
    KeyError,
    IndexError,
    AttributeError,
    AssertionError,
    NotImplementedError,
    StopIteration,
    RuntimeError,
    NameError,
    OSError,
    IOError,
    EOFError,
    ImportError,
    LookupError,
    MemoryError,
    ReferenceError,
    SystemError,
    SystemExit,
    ConnectionError,
    BrokenPipeError,
    ConnectionAbortedError,
    ConnectionRefusedError,
    ConnectionResetError,
    FileExistsError,
    FileNotFoundError,
    InterruptedError,
    IsADirectoryError,
    NotADirectoryError,
    PermissionError,
    ProcessLookupError,
    TimeoutError,
    BlockingIOError,
    ChildProcessError,
    ConnectionError,
    BrokenPipeError,
)



import asyncio
from typing import (
    Callable,
    Any,
    Awaitable,
    Union,
    List,
    Dict,
    Optional,
    Type,
    TypeVar,
    cast,
    overload,
)
from aiohttp.web import (
    Application,
    Request,
    Response,
    json_response,
    AppRunner,
    TCPSite,
)
from aiohttp.web_request import FileField
from aiofauna.odm import AsyncFaunaModel as BaseModel
from inspect import signature, isclass


def create_open_api_params():
    return {"title": "Aiofauna", "version": "0.0.1", "openapi": "3.0.0", "paths": {}}


def update_open_api(
    open_api: Dict[str, Any],
    path: str,
    method: str,
    func: Any,
    open_api_params: Dict[str, Any],
) -> None:
    # Ensure the required keys are present in the open_api dictionary"
    open_api["paths"].setdefault(path, {})[method.lower()] = {
        "summary": func.__name__,
        "description": func.__doc__,
        "parameters": list(open_api_params.values()),
        "responses": {"200": {"description": "OK"}},
    }


async def extract_parameters(request: Request, params: dict):
    args_to_apply = {}
    open_api_params = {}

    for name, param in params.items():
        type_ = param.annotation

        if type_ in (str, int, float, bool) and name in request.match_info:
            open_api_params[name] = {
                "in": "path",
                "name": name,
                "required": True,
                "schema": {"type": type_, "default": param.default, "required": True},
            }
            args_to_apply[name] = request.match_info[name]

        elif type_ in (str, int, float, bool) and name in request.query:
            open_api_params[name] = {
                "in": "query",
                "name": name,
                "required": True,
                "schema": {"type": type_, "default": param.default, "required": True},
            }
            args_to_apply[name] = type_(request.query[name])

        elif issubclass(type_, BaseModel):
            open_api_params[name] = {
                "in": "body",
                "name": name,
                "required": True,
                "schema": {
                    "type": "object",
                    "default": param.default,
                    "required": True,
                },
            }
            args_to_apply[name] = type_(**await request.json())

        elif issubclass(type_, FileField):
            open_api_params[name] = {
                "in": "formData",
                "name": name,
                "required": True,
                "schema": {"type": "file", "default": param.default, "required": True},
            }
            args_to_apply[name] = await request.post()

        elif issubclass(type_, Request):

            continue

        else:

            continue

    return args_to_apply, open_api_params


class Api(Application):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.openapi = create_open_api_params()

    def document(self, path: str, method: str):
        def decorator(func):
            async def wrapper(*args, **kwargs) -> Response:
                sig = signature(func)
                params = sig.parameters
                request: Request = args[0]
                args = args[1:]
                args_to_apply, open_api_params = await extract_parameters(
                    request, params.copy()
                )

                update_open_api(self.openapi, path, method, func, open_api_params)
                if asyncio.iscoroutinefunction(func):
                    return await func(*args, **args_to_apply, **kwargs)
                return func(*args, **args_to_apply, **kwargs)

            return wrapper

        return decorator

    def get(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_get(path, self.document(path, "GET")(func), **kwargs)
            return func

        return decorator

    def post(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_post(path, self.document(path, "POST")(func), **kwargs)
            return func

        return decorator

    def put(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_put(path, self.document(path, "PUT")(func), **kwargs)
            return func

        return decorator

    def delete(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_delete(path, self.document(path, "DELETE")(func), **kwargs)
            return func

        return decorator

    def patch(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_patch(path, self.document(path, "PATCH")(func), **kwargs)
            return func

        return decorator

    def head(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_head(path, self.document(path, "HEAD")(func), **kwargs)
            return func

        return decorator

    def options(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_options(
                path, self.document(path, "OPTIONS")(func), **kwargs
            )
            return func

        return decorator

    def static(self, path: str, directory: str, **kwargs):
        def decorator(func):
            self.router.add_static(path, directory, **kwargs)
            return func

        return decorator

    async def docs(self, request: Request) -> Response:
        return json_response(self.openapi)

    async def listen(self, host: str = "0.0.0.0", port: int = 8000, **kwargs):
        
        
        for route in list(self.router.routes()):
            handler = route.handler.__closure__[0].cell_contents
            for k, v in signature(handler).parameters.items():
                if isclass(v.annotation) and issubclass(v.annotation, BaseModel):
                    self.openapi["paths"][route.url_for.__path][route.method.lower()][
                        "parameters"
                    ].append(
                        {
                            "in": "body",
                            "name": k,
                            "required": True,
                            "schema": v.annotation.schema(),
                        }
                    )

                elif v in (str, int, float, bool):
                    self.openapi["paths"][route.url_for.__path][route.method.lower()][
                        "parameters"
                    ].append(
                        {
                            "in": "query",
                            "name": k,
                            "required": True,
                            "schema": {
                                "type": v.annotation,
                                "default": v.default,
                                "required": True,
                            },
                        }
                    )

                elif issubclass(v.annotation, FileField):

                    self.openapi["paths"][route.url_for.__path][route.method.lower()][
                        "parameters"
                    ].append(
                        {
                            "in": "formData",
                            "name": k,
                            "required": True,
                            "schema": {
                                "type": "file",
                                "default": v.default,
                                "required": True,
                            },
                        }
                    )

                elif issubclass(v.annotation, Request):

                    continue

                else:

                    continue


                

        self.router.add_get("/docs", self.document("/docs", "GET")(self.docs))

        runner = AppRunner(self)

        await runner.setup()

        site = TCPSite(runner, host, port)

        await site.start()

        while True:

            await asyncio.sleep(3600)
